@@ -22,12 +22,13 @@ from enum import Enum
 import json 
 import os
 from pathlib import Path
+import re
 import shutil
 import threading
 
 from PIL import UnidentifiedImageError
-import torch
 import skorch
+import torch
 
 from experiment_manager.neural_net_config import NeuralNetConfig, ConfigError
 import matplotlib.pyplot as plt
@@ -38,7 +39,7 @@ import torch.nn as nn
 
 class Datatype(Enum):
     tabular 	= '.csv'
-    model   	= '.pth'
+    model   	= ['.pth', '.pkl']
     figure  	= '.pdf'
     hparams     = '.json'
     tensorboard = 'tensorboard'
@@ -436,7 +437,7 @@ TODO:
     # read
     #-----
     
-    def read(self, key, datatype, initialized_net=None):
+    def read(self, key, datatype, uninitialized_net=None):
         '''
         Given the key used in a previous save()
         call, and the datatype (Datatype.tabular, 
@@ -470,9 +471,9 @@ TODO:
             a table (i.e. csv file), a figure (.pdf/.png, etc),
             or one of the other Datatype enums
         :type datatype: Datatype
-        :param initialized_net: initialized neural net; only
+        :param uninitialized_net: initialized neural net; only
             needed for models 
-        :type initialized_net: {skorch.classifier.NeuralNet | torch.nn.Module
+        :type uninitialized_net: {skorch.classifier.NeuralNet | torch.nn.Module
         :returns retrieved item
         :rtype {any}
         '''
@@ -508,7 +509,7 @@ TODO:
             # they are all pytorch state_dict exports.
             # But for skorch we also have optimizer and
             # history state, so loading is different:
-            return self._load_model(path, initialized_net)
+            return self._load_model(path, uninitialized_net)
 
         elif datatype == Datatype.figure:
             try:
@@ -584,7 +585,18 @@ TODO:
         if datatype == Datatype.tabular:
             path = os.path.join(self.csv_files_path, f"{key}{dt_extension}")
         elif datatype == Datatype.model:
-            path = os.path.join(self.models_path, f"{key}{dt_extension}")
+            # Models may have one of two extensions:
+            # .pkl for skorch, or .pth for pytorch. Also,
+            # for skorch models the models directory will
+            # have an opt.pkl, in addtion to the file
+            # whose name matches the key:
+            path = None
+            for model_file in os.listdir(self.models_path):
+                file_p = Path(model_file)
+                if file_p.suffix in dt_extension and \
+                   file_p.stem == key:
+                    path = os.path.join(self.models_path, model_file)
+                    break
         elif datatype == Datatype.figure:
             # Figures may have different extensions: png, pdf, etc.:
             path = None
@@ -639,6 +651,60 @@ TODO:
             raise ValueError("Saftey passphrase is not 'Yes, do it', so experiment not cleared")
         shutil.rmtree(self.root)
 
+    #------------------------------------
+    # collect_experiment_roots
+    #-------------------
+    
+    @classmethod
+    def collect_experiment_roots(cls, all_exps_root, common_time_stamp):
+        '''
+        Given the root of potentially many experiment directories, and
+        the timestamp under which some subset were created from running
+        the run_inference.py script, return a dict mapping species to 
+        experiment directory name.
+        
+        The directories will be just the names without the full
+        path, which can be recreated easily by prependeing all_exps_root.
+        
+        Experiment roots look like this:
+        
+              Classifier_WTROC_2021-09-18T15_00_16_inference
+
+        :param all_exps_root: root of a set of experiment manager
+            root subdirectories
+        :type all_exps_root: str
+        :return: dict mapping each species to the name of 
+            an inference subdirectory; may be empty
+        :rtype: {str : str}
+        '''
+        species_date_pattern = re.compile(r"Classifier_([A-Z]{5})_([T0-9-_]{19})_inference")
+
+        dirs = []
+        for exp_dir_name in os.listdir(all_exps_root):
+            exp_dir = os.path.join(all_exps_root, exp_dir_name)
+            dir_name = Path(exp_dir).stem
+            if not os.path.isdir(exp_dir) or \
+               not dir_name.startswith('Classifier') or \
+               not dir_name.endswith('_inference'):
+                continue
+            match = species_date_pattern.search(dir_name)
+            if match is None:
+                # Some other type of directory:
+                continue
+            species, timestamp = match.groups()
+            if timestamp != common_time_stamp:
+                # It's an experiment from a differenct
+                # training:
+                continue
+            
+            # Remember the species and the dir name
+            # of the experiment (without parents):
+            dirs.append((species, dir_name))
+            
+        # Sort by species:
+        exp_dict = dict(sorted(dirs))
+        return exp_dict
+
 
     # --------------- Private Methods --------------------
 
@@ -691,7 +757,7 @@ TODO:
                 # of .pth. Correct the extension:
                 net_path_p = net_path_p.parent.joinpath(net_path_p.stem + '.pkl')
                 
-                optimizer_path = net_path_p.parent.joinpath('optimizer.pkl')
+                optimizer_path = net_path_p.parent.joinpath('opt.pkl')
                 history_path   = net_path_p.parent.joinpath('history.json')
                 net.load_params(
                     f_params=net_path_p, 
@@ -703,8 +769,9 @@ TODO:
                     net.load_params(
                         f_params=path, 
                         f_optimizer=str(optimizer_path),
-                        f_history=str(history_path),
-                        map_location=torch.device('cpu')
+                        f_history=str(history_path)
+                        #******* FIND HOW TO RUN NET ON DIFFERENT DEVICE
+                        #f_map_location=torch.device('cpu')
                         )
         else:
             raise TypeError(f"The network passed into read() must be a pytorch or skorch net, not {type(net)}")
@@ -777,7 +844,12 @@ TODO:
             if Path(path).suffix == '.csv':
                 # Get the field names (i.e. header row):
                 with open(path, 'r') as fd:
-                    col_names = csv.DictReader(fd).fieldnames
+                    # Macos sometimes adds weird quarantine files
+                    # with a leading underscore; skip those:
+                    try:
+                        col_names = csv.DictReader(fd).fieldnames
+                    except UnicodeDecodeError as _e:
+                        continue
             
                 # Make the writer:
                 fd = open(path, 'a')
@@ -898,8 +970,13 @@ TODO:
         except KeyError:
             # No CSV writer yet:
             if header is None:
-                header = self._get_field_names(item, index_col=index_col, trust_list_dim=trust_list_dim)
-
+                header = self._get_field_names(item, 
+                                               index_col=index_col, 
+                                               trust_list_dim=trust_list_dim)
+            else:
+                # Is the index_col header names provided?
+                if index_col is not None:
+                    header = [index_col] + header
             fd = open(dst, 'w')
             csv_writer = csv.DictWriter(fd, header)
             # Save the fd with the writer obj so
@@ -920,7 +997,8 @@ TODO:
                 if len(header) == len(col_names) + 1:
                     index_col = header[0]
                 else:
-                    index_col = None
+                    # Used to be "index_col = None"
+                    index_col = index_col
 
         # Now the DictWriter exists; write the data.
         # Method for writing may vary with data type.
@@ -1402,6 +1480,22 @@ TODO:
         super().__delitem__(key)
         self._schedule_save()
 
+    #------------------------------------
+    # __repr__
+    #-------------------
+    
+    def __repr__(self):
+        root_dir_name = Path(self.root).name
+        return f"<ExpMan {root_dir_name} {hex(id(self))}"
+
+    #------------------------------------
+    # __hash__ 
+    #-------------------
+    
+    def __hash__(self):
+        return id(self)
+        
+
 # ------------------- Class AutoSaveThread -----------------
 
 class AutoSaveThread(threading.Thread):
